# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from __future__ import absolute_import
import pytest
from mercurial import (
    error,
    phases,
)
from heptapod.testhelpers import (
    LocalRepoWrapper,
    )

from .utils import switch_user


def init_repo(basedir):
    return LocalRepoWrapper.init(
        basedir,
        config=dict(
            phases={'publish': 'no'},
            hooks={'pretxnclose.check_publish':
                   'python:heptapod.hooks.check_publish.check_publish'
                   },
            web={'allow-publish': 'maintainer',
                 'allow-push': '*',
                 },
        ))


def test_draft_publish(tmpdir):
    wrapper = init_repo(tmpdir)
    switch_user(wrapper, 'someone')
    ctx = wrapper.write_commit('foo', content='Foo', return_ctx=True)
    assert ctx.phase() == phases.draft

    with pytest.raises(error.Abort) as exc_info:
        wrapper.set_phase('public', [ctx.hex()])
    expected_msg = 'user "someone" is not authorised to publish changesets'
    assert expected_msg in exc_info.value.args[0]

    switch_user(wrapper, 'maintainer')
    wrapper.set_phase('public', [ctx.hex()])
    assert ctx.phase() == phases.public


def test_wrong_hook(tmpdir):
    wrapper = init_repo(tmpdir)
    ui = wrapper.repo.ui
    pretxn = 'pretxnclose.check_publish'
    hookdef = ui.config('hooks', pretxn)
    ui.setconfig('hooks', pretxn, '')
    ui.setconfig('hooks', 'precommit.check_publish', hookdef)
    # precommit because that one does not swallow exceptions other
    # than abort
    with pytest.raises(error.ProgrammingError) as exc_info:
        wrapper.write_commit('foo')

    assert 'precommit' in exc_info.value.args[0]



# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from __future__ import absolute_import

from mercurial import (
    error,
    scmutil,
)
import pytest

from heptapod.testhelpers import (
    LocalRepoWrapper,
)
from .utils import common_config


def test_hpd_unique_successor(tmpdir, monkeypatch):
    repo_path = tmpdir.join('repo')
    wrapper = LocalRepoWrapper.init(repo_path, config=common_config())
    ctx = wrapper.write_commit('foo', message="default0",
                               return_ctx=True)
    repo_path.join('foo').write('amend 1')
    wrapper.command('amend', message='amend1')
    repo_path.join('foo').write('amend 2')
    wrapper.command('amend', message='amend2')

    records = []

    def write(*args, **opts):
        records.append((args, opts))

    wrapper.repo.ui.write = write
    wrapper.command('hpd-unique-successor', rev=ctx.hex())
    out = records[0][0][0]

    succ_ctx = scmutil.revsingle(wrapper.repo, out)
    assert succ_ctx.description() == 'amend2'


def test_hpd_unique_successor_divergence(tmpdir, monkeypatch):
    repo_path = tmpdir.join('repo')
    config = common_config()
    config.setdefault('experimental', {})['evolution.allowdivergence'] = 'yes'
    wrapper = LocalRepoWrapper.init(repo_path, config=config)
    ctx = wrapper.write_commit('foo', message="default0",
                               return_ctx=True)
    repo_path.join('foo').write('amend 1')
    wrapper.command('amend', message='amend1')

    # let's create the divergence
    wrapper.update(ctx.hex(), hidden=True)
    repo_path.join('foo').write('amend 2')
    wrapper.command('amend', message='amend2')

    with pytest.raises(error.Abort) as exc_info:
        wrapper.command('hpd-unique-successor', rev=ctx.hex())
    assert 'divergent' in exc_info.value.args[0]


def test_hpd_unique_successor_missing_rev(tmpdir, monkeypatch):
    repo_path = tmpdir.join('repo')
    wrapper = LocalRepoWrapper.init(repo_path, config=common_config())

    with pytest.raises(error.Abort) as exc_info:
        wrapper.command('hpd-unique-successor')
    assert 'specify a revision' in exc_info.value.args[0]



# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import os
import pytest
from mercurial import (
    error,
    )
from mercurial.hgweb.common import (
    ErrorResponse,
    )

from ..wsgi import HgServe
from ..testhelpers import (
    LocalRepoWrapper,
    make_ui,
)


def test_missing_repositories_root(tmpdir):
    hgrc_path = tmpdir.join('heptapod.hgrc')
    hgrc_path.write('\n'.join(("[extensions]", "evolve=", "topic=", "")))

    with pytest.raises(ValueError) as exc_info:
        HgServe(conf_path=str(hgrc_path))

    assert 'heptapod.repositories-root' in exc_info.value.args[0]


def test_missing_conf_file(tmpdir):
    hgrc_path = tmpdir.join('heptapod.hgrc')
    hgrc_path.write('\n'.join(("[extensions]", "evolve=", "topic=", "")))

    with pytest.raises(error.Abort) as exc_info:
        HgServe(conf_path=str(tmpdir.join('does-not-exist')))

    assert 'not found' in exc_info.value.args[0]


def test_missing_secondary_conf_file(tmpdir):
    """Only the first HGRC in path is required."""
    hgrc = tmpdir.join('hgrc')
    repos = tmpdir.join('repositories')
    hgrc.write('\n'.join(('[heptapod]\n',
                          'repositories-root=' + str(repos),
                          )))
    hgs = HgServe(conf_path=':'.join((
        str(hgrc),
        str(tmpdir.join('does-not-exist'))
    )))
    assert hgs.repos_root == str(repos)


def make_hgserve(repo_root):
    """Return a minimal instance of Hgserve for given repository root."""
    base_ui = make_ui(None, config=dict(
        heptapod={'repositories-root': str(repo_root)}
    ))
    return HgServe(baseui=base_ui)


def test_repositories_root(tmpdir):
    hgs = make_hgserve(tmpdir)
    assert hgs.repos_root == str(tmpdir)


def test_config_from_environ(tmpdir):
    assert 'HEPTAPOD_HGRC' not in os.environ  # checking our premises
    hgrc = tmpdir.join('hgrc')
    repos = tmpdir.join('repositories')
    hgrc.write('\n'.join(('[heptapod]\n',
                          'repositories-root=' + str(repos),
                          )))
    try:
        os.environ['HEPTAPOD_HGRC'] = str(hgrc)
        hgs = HgServe.default_app()
        assert hgs.repos_root == str(repos)
    finally:
        del os.environ['HEPTAPOD_HGRC']

    # really make sure we won't pollute other tests
    assert 'HEPTAPOD_HGRC' not in os.environ


def test_apply_heptapod_headers(tmpdir):
    hgs = make_hgserve(tmpdir)

    env = {'HTTP_X_HEPTAPOD_PERMISSION_USER': 'permuser',
           'REMOTE_ADDR': '::1'}
    hgs.apply_heptapod_headers(env)
    assert env.get('REMOTE_USER') == 'permuser'

    env = {'HTTP_X_HEPTAPOD_PERMISSION_USER': 'permuser',
           'REMOTE_ADDR': '10.1.2.3'}
    hgs.apply_heptapod_headers(env)
    # TODO we might want to revise this and put an explicit denying user
    # in there, since at least in check_publish, absence of REMOTE_USER
    # is interpreted as running from the command line
    assert 'REMOTE_USER' not in env

    env = {'HTTP_X_HEPTAPOD_PROJECT_ID': '18'}
    hgs.apply_heptapod_headers(env)
    assert env.get('HEPTAPOD_PROJECT_ID') == '18'


def test_load_repo(tmpdir):
    wrapper = LocalRepoWrapper.init(
        tmpdir.join('group').ensure(dir=True)
        .join('proj.hg'))
    wrapper.write_commit('foo', message='Foo')
    hgs = make_hgserve(tmpdir)
    hgs.ui.setconfig('heptapod', 'canary', 'yellow')

    repo = hgs.load_repo('group/proj.hg')
    assert repo[0].description() == 'Foo'

    repo.ui.setconfig('heptapod', 'canary', 'red')
    assert repo.ui.config('heptapod', 'canary') == 'red'
    assert hgs.ui.config('heptapod', 'canary') == 'yellow'


def test_load_repo_not_found(tmpdir):
    hgs = make_hgserve(tmpdir)
    with pytest.raises(ErrorResponse) as exc_info:
        hgs.load_repo('does/not/exist')

    assert exc_info.value.args[0] == 'Not Found'



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Interaction with Git repos.

It is not expected that the rise of the HGitaly project would remove
all interesting things to do with Git.
"""
from dulwich.protocol import ZERO_SHA
from dulwich.repo import check_ref_format
from hggit.git_handler import GitHandler
from heptapod import gitlab
from hgext3rd.evolve import headchecking
from mercurial.i18n import _
from mercurial.node import hex
from mercurial import (
    error,
    hg,
    phases,
    pycompat,
)
import re
import traceback

from . import obsutil


class MultipleDescendants(LookupError):
    pass


class InvalidRef(ValueError):
    pass


def git_branch_ref(branch):
    return 'refs/heads/' + branch


TOPIC_REF_PREFIX = git_branch_ref('topic/')

NAMED_BRANCH_REF_PREFIX = git_branch_ref('branch/')


def ref_is_topic(ref):
    return ref.startswith(TOPIC_REF_PREFIX)


def ref_is_named_branch(ref):
    return ref.startswith(NAMED_BRANCH_REF_PREFIX)


def parse_git_branch_ref(ref):
    """Parse a Git branch ref for named branch and topic information

    Return ``None`` if ``ref`` is not a Git branch. This is considered
    to be a normal case::

      >>> parse_git_branch_ref("refs/tags/v1.2.3") is None
      True

    Example with a topical GitLab branch::

      >>> parse_git_branch_ref("refs/heads/topic/default/thetop")
      ('default', 'thetop')

    For named branches, the returned topic is ``None``. Checking for `None`
    is the normal way for downstream code to know if this is a topic or not::

      >>> parse_git_branch_ref("refs/heads/branch/default")
      ('default', None)

    Special case: forward slashes are considered part of the named branch::

      >>> parse_git_branch_ref("refs/heads/branch/ze/branch")
      ('ze/branch', None)

      >>> try:
      ...     _ = parse_git_branch_ref("refs/heads/topic/invalid")
      ... except InvalidRef as exc:
      ...     print(exc.args[0])
      ... else:
      ...     print("Expected InvalidRef exception not raised")
      refs/heads/topic/invalid
    """
    git_ref_prefix = 'refs/heads/'
    if not ref.startswith(git_ref_prefix):
        return None
    git_branch = ref[len(git_ref_prefix):]

    topic_prefix = 'topic/'
    if git_branch.startswith(topic_prefix):
        res = tuple(git_branch.split('/', 1)[1].rsplit('/', 1))
        if len(res) != 2:
            raise InvalidRef(ref)
        return res

    named_branch_prefix = 'branch/'
    if git_branch.startswith(named_branch_prefix):
        return git_branch[len(named_branch_prefix):], None


class GitRefChange(object):
    """Represent a change to be performed in a target Git repo.

    Public attributes:

    - :attr:`ref`: Git ref name
    - :attr:`before`, :attr:`after`: Git SHAs

    These will be complemented with a system of options, e.g., to specify that
    a topic change actually comes with publication, leading to a deferred
    removal of the corresponding Git branch once all appropriate treatments are
    done, whether this removal is performed from here or by GitLab.
    """

    def __init__(self, ref, before, after):
        self.ref = ref
        self.before = before
        self.after = after

    def is_creation(self):
        return self.before == ZERO_SHA and self.after != ZERO_SHA


class HeptapodGitHandler(GitHandler):

    def __init__(self, *args, **kwargs):
        super(HeptapodGitHandler, self).__init__(*args, **kwargs)

        main_repo = hg.sharedreposource(self.repo)
        if main_repo is None:
            main_repo = self.repo

        self.gitdir = re.sub(r'\.hg$', '', main_repo.root) + '.git'
        self._gl_hooks = {}
        self.unfiltered_repo = self.repo.unfiltered()

    def heptapod_gate_bookmarks(self, repo, allow, changes):
        """First handling of bookmark changes (refuse or not), return deleted.

        :param repo: passed explicitely because filtering may differ from
                     :attr:`repo`.
        :return: iterable of deleted bookmarks
        """
        if not changes:
            return ()

        ui = repo.ui
        deleted = []
        ui.note("HeptapodGitHandler bookmark changes=%r" % changes)
        new_bookmarks = []

        for name, change in changes.items():
            if change[0] is None:
                new_bookmarks.append(name)
            elif change[1] is None:
                deleted.append(name)

        if new_bookmarks and not allow:
            raise error.Abort(_(
                "Creating bookmarks is forbidden by default in Heptapod "
                "(got these new ones: %r). "
                "See https://heptapod.net/pages/faq.html#bookmarks to "
                "learn why and how to partially lift "
                "that restriction" % new_bookmarks))
        for new_bm_name, (_ign, new_bm_node) in pycompat.iteritems(changes):
            new_bm_ctx = repo[new_bm_node]
            new_bm_topic = new_bm_ctx.topic()
            if new_bm_topic:
                raise error.Abort(_(
                    "Creating or updating bookmarks on topic "
                    "changesets is forbidden"),
                    hint="topic %r and bookmark %r for changeset %s" % (
                        new_bm_topic, new_bm_name, new_bm_ctx.hex()))
        return deleted

    def get_exportable(self):
        """Heptapod version, including named branches and topics.

        This rewraps :meth:`GitHandler.get_exportable` to add named branches
        and topics to the returned Git refs
        """
        git_refs = super(HeptapodGitHandler, self).get_exportable()
        logprefix = 'HeptapodGitHandler.get_exportable '
        repo = self.repo.filtered('served')
        ui = repo.ui

        # "exporting" the ZERO_SHA will mean by convention a request to prune
        # the corresponding heads -- only a request, we can't really decide
        # at this stage.
        # The value is a dict mapping refs to reasons (just strings)
        to_prune = git_refs[ZERO_SHA] = {}

        txn = repo.currenttransaction()
        allow_bookmarks = ui.configbool('experimental',
                                        'hg-git.bookmarks-on-named-branches')
        bm_changes = txn.changes.get('bookmarks') if txn is not None else None
        deleted_bms = self.heptapod_gate_bookmarks(
            repo, allow_bookmarks, bm_changes)

        to_prune.update((git_branch_ref(bm), 'deleted-bookmark')
                        for bm in deleted_bms)
        all_bookmarks = self.repo._bookmarks
        # currently, self_filter_for_bookmarks() only does some renaming,
        # and doesn't discard any, so this is actually just equivalent to
        # hgshas of all bookmarks, but that may change in future hg-git
        hgshas_with_bookmark_git_ref = {
            hex(all_bookmarks[bm])
            for _, bm in self._filter_for_bookmarks(all_bookmarks)}
        for branch, hg_nodes in repo.branchmap().iteritems():
            gb = self.git_branch_for_branchmap_branch(branch)
            revs = [repo[n].rev() for n in hg_nodes]
            ctxs = [repo[r]
                    for r in headchecking._filter_obsolete_heads(repo, revs)]

            if ui.configbool('experimental',
                             'hg-git.prune-newly-closed-branches', True):
                ctxs = [c for c in ctxs if not c.closesbranch()]
                if not ctxs:
                    to_prune[git_branch_ref(gb)] = 'closed'

            hg_shas = {c.hex() for c in ctxs}
            # We ignore bookmarked changesets because:
            # - we don't want them in the 'wild' namespace
            # - no other Git ref would be relevant for them
            hg_shas.difference_update(hgshas_with_bookmark_git_ref)
            if not hg_shas:
                if hg_nodes and allow_bookmarks:
                    # after removal of bookmarks, but not before,
                    # there is no hg head on this named branch:
                    # schedule potential removal. We don't want to do
                    # this if bookmarks aren't explicitely allowed because
                    # this must be a side effect, potentially disturbing, of
                    # an implicit bookmark move
                    to_prune[git_branch_ref(gb)] = 'only-bookmarks'
                ui.note(logprefix,
                        "Branch %r has no visible"
                        "non-bookmarked head" % branch)
                continue

            if 1 < len(hg_shas):
                ui.note(logprefix, "Multiple heads for branch %r: %r" % (
                    branch, hg_shas))
                for hg_sha in hg_shas:
                    git_refs[hg_sha].heads.add(
                        git_branch_ref('wild/' + hg_sha))
                gca = self.multiple_heads_choose(hg_shas, branch)
                if gca is None:
                    # giving up in order to avoid confusing situations
                    continue
                gca_sha = self.repo[gca].hex()
                ui.note(logprefix,
                        "Chose %r out of multiple heads %r "
                        "for forwarding branch %r" % (
                            gca_sha, hg_shas, branch))
                hg_shas = [gca_sha]
            hg_sha = hg_shas.pop()
            gb = self.git_branch_for_branchmap_branch(branch)
            git_refs[hg_sha].heads.add(git_branch_ref(gb))
        return git_refs

    def published_topic_latest_hg_sha(self, topic, ctx, log_before_ctx=None):
        """Rewrapping of `latest_topic_descendant` returning hg sha.
        """
        ui = self.repo.ui
        try:
            after_ctx = self.latest_topic_descendant(topic, ctx)
        except MultipleDescendants:
            msg = ("Found several descendants in topic %r of the "
                   "newly published %r. Can't have GitLab add them "
                   "to any related merge request. ") % (topic, ctx)
            ui.warn(msg)
            ui.status(msg)
            # since successor is public, chances to detect merge are good
            return ctx.hex()

        if after_ctx is None:
            ui.warn("HeptapodGitHandler.published_topic_latest_hg_sha "
                    "inspecting public changeset %r for topic %r gave "
                    "inconsistent result: it's not in the expected topic. "
                    "This will trigger immediate pruning of the "
                    "topic Git branch" % (ctx, topic))
            return None

        ui.note('HeptapodGitHandler.published_topic_latest_hg_sha',
                "updating published %r from %r to %r" % (
                    topic,
                    log_before_ctx if log_before_ctx is not None else ctx,
                    after_ctx))
        return after_ctx.hex()

    def analyse_vanished_topic(self, branch, topic, before_sha,
                               log_info):
        """Compute revision to send GitLab as new topic branch head.

        This method does not access the Git repository.

        :param before_sha: Mercurial SHA for the current Git head of the topic
                           GitLab branch. Can be ``None`` if the Git head is
                           actually unknown to Mercurial.
        :param log_info: dict of useful information for logs that we must
                         restrain not to use for other purposes in this method
                         (typically some Git context)
        :return: Mercurial SHA, or ``None`` to trigger pruning.
        """
        logprefix = 'HeptapodGitHandler.topic_new_hg_sha '
        ui = self.repo.ui
        initial_import = ui.configbool(b"heptapod", b"initial-import")
        try:
            # TODO case where before_sha is None. same error treatment?
            before_ctx = self.unfiltered_repo[before_sha]
        except error.RepoLookupError:
            ui.warn(logprefix, "Git {ref!r} "
                    "(gitsha={before_git_sha}, hgsha={before_sha}) not "
                    "found in the Mercurial repo (should be due to "
                    "some half-rollbacked previous transaction), pruning "
                    "as the topic does not seem to be visible anymore".format(
                        before_sha=before_sha, **log_info))
            return None

        if before_ctx.phase() == phases.public:
            latest_sha = self.published_topic_latest_hg_sha(topic, before_ctx)
            if latest_sha is not None:
                return latest_sha
            else:
                # this is a corruption: resolving the Git branch for the topic
                # actually gives us a changeset that does not bear that topic!
                # This is what happened in heptapod#265.
                # In that case, we return it unchanged. The changeset
                # surely is an ancestor of the current named branch head
                # GitLab's MR detection should thus work and eventually
                # prune the corrupted Git branch for the topic.
                return before_sha
        try:
            succctx = obsutil.latest_unique_successor(before_ctx)
        except error.Abort as exc:
            if initial_import:
                # we don't want to break an initial import because
                # of an exceptional phase divergence, let's keep it unchanged
                ui.warn(logprefix + exc.args[0])
                return before_sha
            else:
                raise

        if succctx is None:
            ui.note(logprefix,
                    "scheduling prune of {ref} "
                    "(obsolete, no successor)".format(**log_info))
            return None

        succ_phase = succctx.phase()
        if succ_phase == phases.public:
            if not initial_import:
                latest_sha = self.published_topic_latest_hg_sha(
                    topic, succctx, log_before_ctx=before_ctx)
                if latest_sha is not None:
                    return latest_sha

        elif succ_phase == phases.draft:
            # let's go over some reasons why there's no visible branch/topic
            # head and the former one is obsolete with a draft successor
            succ_brtop = (succctx.branch(), succctx.topic())
            before_brtop = (before_ctx.branch(), before_ctx.topic())
            if succ_brtop != before_brtop:
                ui.note(logprefix,
                        "pruning %r (hgsha {before}), as its successor {succ} "
                        "is on another branch or topic".format(
                            before=before_ctx.hex(),
                            succ=succctx.hex(),
                            **log_info))
                return None

        return succctx.hex()

    def generate_prune_changes(self, to_prune, existing):
        """Generate those pruning Git changes that really have to be done.

        :return: a `dict` mapping Git refs to the :class:`Changes`
                 instances that will prune them.

        This method does not use the Git repo, only the :meth:`map_hg_get`
        """
        changes = {}
        prune_previously_closed_branches = self.repo.ui.configbool(
            'experimental', 'hg-git.prune-previously-closed-branches')
        for ref, reason in pycompat.iteritems(to_prune):
            before_sha = existing.get(ref)
            if before_sha is None:
                continue
            if reason == 'closed' and not prune_previously_closed_branches:
                before_hg_sha = self.map_hg_get(before_sha)
                if before_hg_sha is not None:
                    try:
                        before_ctx = self.repo[before_hg_sha]
                    except error.RepoLookupError:
                        # we can't really know if the closing is new,
                        # but being unknown to Mercurial, it certainly has
                        # to be pruned
                        self.repo.ui.warn(
                            "Pruning closed branch %r, whose "
                            "latest Git commit %r corresponds to the unknown "
                            "%r hg sha" % (ref, before_sha, before_hg_sha))
                    else:
                        if before_ctx.closesbranch():
                            # it was already closed
                            continue
            changes[ref] = GitRefChange(ref, before_sha, ZERO_SHA)
        return changes

    def analyze_vanished_refs(self, existing, exportable):
        """Decide what to do of existing Git refs that aren't in exportable.

        Some will have to be pruned, some other (published topics) will
        not.

        :return: dict mapping refs to `GitRefChange` instance to apply
        """
        changes = {}
        exported_refs = {ref for heads_tags in exportable.values()
                         for ref in heads_tags}
        for ref, before_git_sha in pycompat.iteritems(existing):
            if ref in exported_refs:
                continue

            if ref.startswith('refs/heads/wild/'):
                changes[ref] = GitRefChange(ref, before_git_sha, ZERO_SHA)

            try:
                branch_topic = parse_git_branch_ref(ref)
            except InvalidRef:
                self.repo.ui.warn("Git mirror repo has a bogus branch ref %r "
                                  "that's not among the exportable ones. "
                                  "Ignoring it.")
                continue
            if branch_topic is not None and branch_topic[1]:
                branch, topic = branch_topic
                before_hg_sha = self.map_hg_get(before_git_sha)
                after_hg_sha = self.analyse_vanished_topic(
                    branch, topic, before_hg_sha,
                    log_info=dict(ref=ref, before_git_sha=before_git_sha),
                    )
                if after_hg_sha is None:
                    after_git_sha = ZERO_SHA
                else:
                    after_git_sha = self.map_git_get(after_hg_sha)
                    if after_git_sha is None:
                        self.repo.ui.warn(
                            "Analysis of topic %r for branch %r "
                            "that becomes invisible in this transaction "
                            "to report its latest change to GitLab "
                            "found Mercurial changeset %r that has no known "
                            "Git counterpart. Giving up on reporting "
                            "that topic/branch combination to GitLab." % (
                                topic, branch, after_hg_sha))
                        continue

                changes[ref] = GitRefChange(ref, before_git_sha, after_git_sha)
        return changes

    def compare_exportable(self, existing, exportable):
        """Analyse the exportable refs to produce Git changes

        :param exportable: a mapping from Git refs to Mercurial SHAs
        :param existing: a mapping of existing Git refs in the target repo
                         to Git SHAs
        :returns: a mapping from Git refs to :class:`GitRefChange` objects
        """
        # HEAD is the source of truth for GitLab default branch
        default_branch_ref = self.git.refs.get_symrefs().get('HEAD')
        to_prune = exportable.pop(ZERO_SHA, {})
        to_prune.pop(default_branch_ref, None)
        changes = self.generate_prune_changes(to_prune, existing)

        changes.update(self.analyze_vanished_refs(existing, exportable))

        for hg_sha, refs in pycompat.iteritems(exportable):
            for ref in refs.heads:
                after_sha = self.map_git_get(hg_sha)
                before_sha = existing.get(ref, ZERO_SHA)
                if after_sha and after_sha != before_sha:
                    changes[ref] = GitRefChange(ref, before_sha, after_sha)
        return changes

    def gitlab_get_hook(self, name):
        hook = self._gl_hooks.get(name)
        if hook is None and gitlab is not None:
            hook = self._gl_hooks[name] = gitlab.Hook(name, self.repo)
        return hook

    def heptapod_notify_gitlab(self, hook_name, changes, allow_error=False):
        if not changes:
            return

        hook = self.gitlab_get_hook(hook_name)
        ui = self.repo.ui

        try:
            code, out, err = hook({ref: (ch.before, ch.after)
                                   for ref, ch in pycompat.iteritems(changes)})
        except Exception as exc:
            ui.error("GitLab update error (%r hook): %r\n%s" % (
                hook_name, exc, traceback.format_exc()))
            if allow_error:
                # that's an error in py-heptapod, could be a network failure
                # once we call the internal API directly, for instance
                return
            else:
                raise

        if code != 0:
            ui.error("Got code %r while sending GitLab %r hook details=%r" % (
                code, hook_name, err))
            quiet = ui.quiet
            ui.quiet = False
            ui.status("GitLab update error: %r. Because of this, some "
                      "changes won't be visible in the web interface" % err)
            ui.quiet = quiet
            if not allow_error:
                raise error.Abort(err.strip())

        # useful messages such as motd, merge requests links etc.
        quiet = ui.quiet
        ui.quiet = False
        ui.status(out)
        ui.quiet = quiet

    def heptapod_apply_changes(self, changes):
        self.heptapod_notify_gitlab('pre-receive', changes)
        git_refs = self.git.refs

        # Right after repo creation, HEAD is typically initialized
        # as refs/heads/master, which doesn't exist,
        # and probably won't after our push. If we don't correct it
        # quickly, something on the GitLab side, even
        # before post-receive treatment actually begins will set it to
        # a random value - we don't want it to select topics if possible
        # and if it has, we want that to change.
        default_branch_ref = self.git.refs.get_symrefs().get(b'HEAD')
        update_default_branch = (default_branch_ref not in git_refs
                                 or ref_is_topic(default_branch_ref))
        new_named_branch_refs = []
        for change in pycompat.itervalues(changes):
            if change.after == ZERO_SHA:
                del git_refs[change.ref]
            else:
                git_refs[change.ref] = change.after
            if change.is_creation() and ref_is_named_branch(change.ref):
                new_named_branch_refs.append(change.ref)

        if update_default_branch and new_named_branch_refs:
            branch_default = git_branch_ref(b'branch/default')
            if branch_default in new_named_branch_refs:
                new_head = branch_default
            else:
                new_head = new_named_branch_refs[0]
            self.repo.ui.note(b"Setting Git HEAD to %s" % new_head)
            git_refs.set_symbolic_ref(b'HEAD', new_head)

        def post_receive(txn):
            self.heptapod_notify_gitlab('post-receive', changes)

        txn = self.repo.currenttransaction()
        if txn is None:
            post_receive(None)
        else:
            txn.addpostclose('heptapod_git_sync', post_receive)

    def export_commits(self):
        try:
            self.export_git_objects()
            self.update_references()
        finally:
            self.save_map(self.map_file)

    def update_references(self):
        """Update or create refs in the target Git repo for Mercurial changes.

        This fires or schedules all appropriate notifications (GitLab hooks)
        """
        existing = self.git.refs.as_dict()
        changes = self.heptapod_compare_tags(existing)
        changes.update(self.compare_exportable(existing,
                                               self.get_exportable()))
        self.heptapod_apply_changes(changes)

    def heptapod_compare_tags(self, existing):
        """This is derived from export_hg_tags() for two-phase application.

        Instead of immediate application to the Git repo, we emit
        a dict of `GitRefChange` instances, suitable for application between
        pre- and post-receive.
        """
        changes = {}
        for tag, sha in pycompat.iteritems(self.repo.tags()):
            if self.repo.tagtype(tag) in (b'global', b'git'):
                tag = tag.replace(b' ', b'_')
                target = self.map_git_get(hex(sha))
                if target is not None:
                    tag_refname = b'refs/tags/' + tag
                    if check_ref_format(tag_refname):
                        before_sha = existing.get(tag_refname, ZERO_SHA)
                        if before_sha != target:
                            changes[tag_refname] = GitRefChange(
                                tag_refname, before_sha, target)
                    else:
                        self.repo.ui.warn(
                            b"Skipping export of Mercurial tag '%s' because "
                            b"it has invalid name as a git refname.\n" % tag)
                else:
                    self.repo.ui.warn(
                        b"Skipping export of tag '%s' because it "
                        b"has no matching git revision.\n" % tag)
        return changes

    def multiple_heads_choose(self, hg_shas, branchmap_branch):
        """Choose among multiple heads to forward the given branch

        The branch is given in branchmap format, i.e., `branch:topic` or
        `branch`.

        Return None if nothing satisfying has been found.

        Currently, we arbitrarily take the one with the highest revision
        number (hence the most recently added to *this* repository).
        The advantage is that the given branch will never disappear
        (confusing to users, and leading to some blocking situations,
        such as heptapod#101).

        This is consistent with what Mercurial does for label-adressing,
        and shouldn't be a problem for Heptapod, since we force-push to
        Git all the time
        """
        if self.repo.ui.configbool('unit-tests',
                                   'hg-git.multiple-heads-dont-choose'):
            return

        revs = self.repo.revs('max(%ls)', hg_shas)
        if len(revs) != 1:
            return None
        return revs.first()

    def git_branch_for_branchmap_branch(self, topbranch):
        """return Git branch name for hg branch names in branchmap()

        Does in particular the needed sanitizations to make it acceptable
        for a Git branch.

        :param topbranch: can be either a named branch name or follow the
                          branch:topic convention, as returned by branchmap()
        """
        topbranch = topbranch.replace(' ', '_')
        if ':' in topbranch:
            branch, topic = topbranch.split(':', 1)
            if '/' in topic:
                msg = "Invalid character '/' in topic name %r. " % topic
                if self.ui.configbool('experimental',
                                      'hg-git.accept-slash-in-topic-name'):
                    msg += ("Replacing with '-'. Rename before publishing or "
                            "you'll get serious problems.")
                    topic = topic.replace('/', '-')
                    self.ui.status(msg)
                    self.ui.warn(msg)
                else:
                    self.ui.status(msg)
                    raise error.Abort(msg)

            return '/'.join(('topic', branch, topic))
        return 'branch/' + topbranch

    def latest_topic_descendant(self, topic, ctx):
        """Return the latest public descendent of ctx in a given topic.

        Although this is not meaningful in regular Mercurial usage,
        it's necessary because we need GitLab to understand what happened to
        a Merge Request on a published topic, and that's not possible if
        its branch vanished.

        This checks that there is one public head of the topic.
        Otherwise the push should be refused as inconsistent.

        The changeset of ctx is assumed *not* to be filtered (aka, obsolete)
        (all current callers already have obsolescence information around).
        If it does not belong to the topic, `None` gets returned.
        """
        revs = self.repo.revs("heads(extra(topic, %s) and descendants(%d))",
                              topic, ctx.rev())
        if len(revs) > 1:
            raise MultipleDescendants(topic, ctx)
        rev = revs.first()
        if rev is None:
            return None
        return self.repo[rev]



# heptapod/hgweb.py - Heptapod HTTP interface for a directory of repositories.
#
# derived under GPL2+ from Mercurial's hgwebdir_mod.py, whose
# copyright holders are
#   Copyright 21 May 2005 - (c) 2005 Jake Edge <jake@edge2.net>
#   Copyright 2005, 2006 Matt Mackall <mpm@selenic.com>
#
# This file Copyright 2019 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from __future__ import absolute_import

import logging
import gc
import os

from mercurial.i18n import _

from mercurial.hgweb.common import (
    ErrorResponse,
    HTTP_SERVER_ERROR,
    HTTP_NOT_FOUND,
    cspvalues,
    statusmessage,
)

from mercurial import (
    encoding,
    error,
    extensions,
    hg,
    profiling,
    pycompat,
    ui as uimod,
)

from mercurial.hgweb import (
    hgweb_mod,
    request as requestmod,
)

logger = logging.getLogger(__name__)
# logging configuration will be initialized from the Mercurial global
# configuration, overriding this:
logging.basicConfig(level=logging.INFO)

ALLOWED_REMOTES = (
    '127.0.0.1',
    '::1',
    '',  # Unix Domain Socket
    )


# Suffixes of HTTP headers that are directly forwarded in environment
FORWARDED_HEADERS_SUFFIXES = (
    'USERINFO_ID',
    'USERINFO_USERNAME',
    'USERINFO_NAME',
    'USERINFO_EMAIL',
    'PROJECT_ID',
    'PROJECT_PATH',
    'PROJECT_NAMESPACE_FULL_PATH',
    )


class HgServe(object):
    """WSGI application serving repositories under a given root path

    The repositories are expected in the `heptapod.repositories-root`
    Mercurial configuration.

    This works under full trust of the incoming request: callers are either
    `gitlab-rails` or `gitlab-workhorse`.
    """
    def __init__(self, conf_path=None, baseui=None):
        self.baseui = baseui
        if baseui:
            self.ui = baseui.copy()
        else:
            self.ui = uimod.ui.load()
        self.motd = None
        if not baseui:
            # set up environment for new ui
            extensions.loadall(self.ui)
            extensions.populateui(self.ui)
        if conf_path is not None:
            for i, rcf in enumerate(conf_path.split(':')):
                if i == 0 and not os.path.exists(rcf):
                    raise error.Abort(_('config file %s not found!') % rcf)
                logger.info("Loading configuration from %r", rcf)
                self.ui.readconfig(rcf, trust=True)

        root = self.ui.config('heptapod', 'repositories-root')
        if root is None:
            raise ValueError("heptapod.repositories-root is not configured.")
        self.repos_root = root

    def apply_heptapod_headers(self, environ):
        perm_user = environ.get('HTTP_X_HEPTAPOD_PERMISSION_USER')
        if perm_user is not None and environ['REMOTE_ADDR'] in ALLOWED_REMOTES:
            environ['REMOTE_USER'] = perm_user
        for hepta_key in FORWARDED_HEADERS_SUFFIXES:
            hepta_val = environ.pop('HTTP_X_HEPTAPOD_' + hepta_key, None)
            if hepta_val is not None:
                environ['HEPTAPOD_' + hepta_key] = hepta_val

    def __call__(self, env, respond):
        baseurl = self.ui.config('web', 'baseurl')
        self.apply_heptapod_headers(env)
        req = requestmod.parserequestfromenv(env, altbaseurl=baseurl)
        res = requestmod.wsgiresponse(req, respond)
        return self.run_wsgi(req, res)

    def run_wsgi(self, req, res):
        profile = self.ui.configbool('profiling', 'enabled')
        with profiling.profile(self.ui, enabled=profile):
            try:
                for r in self._runwsgi(req, res):
                    yield r
            finally:
                # There are known cycles in localrepository that prevent
                # those objects (and tons of held references) from being
                # collected through normal refcounting. We mitigate those
                # leaks by performing an explicit GC on every request.
                # TODO remove this once leaks are fixed.
                # TODO only run this on requests that create localrepository
                # instances instead of every request.
                gc.collect()

    def load_repo(self, uri_path):
        repo_path = os.path.join(self.repos_root, uri_path)
        if not os.path.isdir(os.path.join(repo_path, '.hg')):
            # hg.repository() would raise a RepoError which is
            # not qualified enough to distinguish it cleanly (just
            # the message)
            raise ErrorResponse(HTTP_NOT_FOUND, "Not Found")
        logger.info("loading repo at %r", repo_path)
        # ensure caller gets private copy of ui
        return hg.repository(self.ui.copy(), repo_path)

    def _runwsgi(self, req, res):
        try:
            csp = cspvalues(self.ui)[0]
            if csp:
                res.headers['Content-Security-Policy'] = csp

            uri_path = req.dispatchpath.strip('/')
            # Re-parse the WSGI environment to take into account our
            # repository path component.
            uenv = req.rawenv
            req = requestmod.parserequestfromenv(
                uenv, reponame=uri_path,
                altbaseurl=self.ui.config('web', 'baseurl'),
                # Reuse wrapped body file object otherwise state
                # tracking can get confused.
                bodyfh=req.bodyfh)
            try:
                repo = self.load_repo(uri_path)
                return hgweb_mod.hgweb(repo).run_wsgi(req, res)
            except IOError as inst:
                msg = encoding.strtolocal(inst.strerror)
                raise ErrorResponse(HTTP_SERVER_ERROR, msg)
            except error.RepoError as inst:  # pragma: no cover
                # This case can't happen because hgweb_mod catches them
                # already. To be removed in Heptapod 0.8
                raise ErrorResponse(HTTP_SERVER_ERROR, bytes(inst))

        except ErrorResponse as e:
            # To be carefully tested with Python3, but Heptapod is running py2
            # for the time being (suitable py3 version not available yet)

            # This is a dubious part from hgweb: the status message includes
            # the error message, but that can get funky because of potential
            # translation. Common practice would be just to repeat the
            # generic meaning of the code, e.g, "NOT FOUND"
            # and provide details in the reponse body
            res.status = statusmessage(e.code, pycompat.bytestr(e))
            res.headers['Content-Type'] = 'text/plain; encoding={}'.format(
                pycompat.sysstr(encoding.encoding))
            res.setbodygen((e.message or b'') + b"\n")
            return res.sendresponse()

    @classmethod
    def default_app(cls):
        HEPTAPOD_HGRC = os.environ.get('HEPTAPOD_HGRC')
        if HEPTAPOD_HGRC:
            return HgServe(conf_path=HEPTAPOD_HGRC)


hgserve = HgServe.default_app()



# Copyright 2019 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Helpers for automatic tests.

These allow both high level operation on testing repos, and lower level
calls and introspections, making it possible to test more exhaustively inner
code paths that with `.t` tests, which are really functional tests.
"""
import os
from mercurial import (
    cmdutil,
    commands,
    hg,
    node,
    phases,
    ui as uimod,
)
import random

# re-exports for stability
NULL_REVISION = node.nullrev
NULL_ID = node.nullid


def make_ui(base_ui, config=None):
    # let's make sure we aren't polluted by surrounding settings
    os.environ['HGRCPATH'] = ''
    if base_ui is None:
        ui = uimod.ui.load()
    else:
        ui = base_ui.copy()
    if ui.environ is os.environ:
        ui.environ = dict(ui.environ)
    if config is not None:
        for section_name, section in config.items():
            for item_name, item_value in section.items():
                ui.setconfig(section_name, item_name, item_value,
                             source='tests')
    return ui


class LocalRepoWrapper(object):

    def __init__(self, repo):
        self.repo = repo

    @classmethod
    def init(cls, path, base_ui=None, config=None):
        path = str(path)
        init = cmdutil.findcmd('init', commands.table)[1][0]
        ui = make_ui(base_ui, config)
        init(ui, dest=path)
        return cls(hg.repository(ui, path))

    @classmethod
    def load(cls, path, base_ui=None, config=None):
        path = str(path)
        ui = make_ui(base_ui, config=config)
        return cls(hg.repository(ui, path))

    @classmethod
    def share_from_path(cls, src_path, dest_path,
                        ui=None, base_ui=None, config=None,
                        **share_opts):
        """Create a new repo as the `share` command would do.

        :param ui: if specified, will be copied and used for the new repo
                   creation through ``share``
        :param config: only if ``ui`` is not specified, will be used to
                       create a new ``ui`` instance
        :param base_ui: only if ``ui`` is not specified, will be used to
                       create a new :class:`ui` instance
        :param share_opts: passed directly to :func:`hg.share()`
        :return: wrapper for the new repo
        """
        if ui is None:
            ui = make_ui(base_ui, config=config)
        else:
            ui = ui.copy()

        # the 'share' command defined by the 'share' extension, is just a thin
        # wrapper around `hg.share()`, which furthermore returns a repo object.
        return cls(hg.share(ui, str(src_path), dest=str(dest_path),
                            **share_opts))

    def share(self, dest_path, **share_opts):
        return self.share_from_path(self.repo.root, dest_path,
                                    ui=self.repo.ui, **share_opts)

    def command(self, name, *args, **kwargs):
        cmd = cmdutil.findcmd(name, commands.table)[1][0]
        repo = self.repo
        return cmd(repo.ui, repo, *args, **kwargs)

    def write_commit(self, rpath,
                     content=None, message=None,
                     return_ctx=False,
                     parent=None,
                     branch=None,
                     topic=None):
        """Write content at rpath and commit in one call.

        This is meant to allow fast and efficient preparation of
        testing repositories. To do so, it goes a bit lower level
        than the actual commit command, so is not suitable to test specific
        commit options, especially if through extensions.

        This leaves the working directoy updated at the new commit.

        :param rpath: relative path from repository root. If existing,
                      will be overwritten by `content`
        :param content: what's to be written in ``rpath``.
                        If not specified, will be replaced by random content.
        :param message: message commit. If not specified, defaults to
                        ``content``
        :param parent: binary node value. If specified, the repository is
                       updated to it first. Useful to produce branching
                       histories. This is single valued, because the purpose
                       of this method is not to produce merge commits.
        :returns: binary node for the resulting commit.
        """
        rpath = str(rpath)
        repo = self.repo
        path = os.path.join(repo.root, rpath)
        if parent is not None:
            self.update_bin(parent)
        if content is None:
            content = "random: {}\n\nparent: {}\n".format(
                random.random(),
                node.hex(repo.dirstate.p1()))
        if message is None:
            message = content

        if branch is not None:
            self.repo.dirstate.setbranch(branch)

        if topic is not None:
            self.command('topics', topic)

        flags = 'wb' if isinstance(content, bytes) else 'w'
        with open(path, flags) as fobj:
            fobj.write(content)

        def commitfun(ui, repo, message, match, opts):
            return self.repo.commit(message,
                                    opts.get('user'),
                                    opts.get('date'),
                                    match=match,
                                    editor=False,
                                    extra=None)
        new_node = cmdutil.commit(repo.ui, repo, commitfun, (path, ),
                                  dict(addremove=True,
                                       message=message))
        return repo[new_node] if return_ctx else new_node

    def update_bin(self, bin_node, **opts):
        """Update to a revision specified by its node in binary form.

        This is separated in order to avoid ambiguities
        """
        # maybe we'll do something lower level later
        self.update(node.hex(bin_node), **opts)

    def update(self, rev, hidden=False):
        repo = self.repo.unfiltered() if hidden else self.repo
        cmdutil.findcmd('update', commands.table)[1][0](repo.ui, repo, rev)

    def set_phase(self, phase_name, revs, force=True):
        repo = self.repo
        opts = dict(force=force, rev=revs)
        opts.update((phn, phn == phase_name) for phn in phases.cmdphasenames)
        cmdutil.findcmd('phase', commands.table)[1][0](repo.ui, repo, **opts)

    def prune(self, revs, successors=(), bookmarks=()):
        # the prune command expects to get all these arguments (it relies
        # on the CLI defaults but doesn't have any at the function call level).
        # They are unconditionally concatened to lists, hence must be lists.
        # (as of Mercurial 5.3.1)
        if isinstance(revs, (bytes, str)):
            revs = [revs]
        return self.command('prune', rev=revs,
                            new=[],  # deprecated yet expected
                            successor=list(successors),
                            bookmark=list(bookmarks))
